from __future__ import annotations

import asyncio
import datetime
import functools
import inspect
import json
import logging
import os
import re
import ssl
import urllib.parse
from dataclasses import dataclass, field
from enum import IntEnum
from importlib.abc import SourceLoader
from importlib.util import module_from_spec, spec_from_loader
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Generic,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
)

import httpx
import kubernetes.client
import yaml
from autogen_core.code_executor import CodeResult
from typing_extensions import ParamSpec
from websockets.asyncio.client import ClientConnection, connect
from websockets.exceptions import ConnectionClosed
from websockets.typing import Subprotocol

T = TypeVar("T")
P = ParamSpec("P")


class StreamChannel(IntEnum):
    STDIN_CHANNEL = 0
    STDOUT_CHANNEL = 1
    STDERR_CHANNEL = 2
    ERROR_CHANNEL = 3
    RESIZE_CHANNEL = 4


class HttpStatusCode(IntEnum):
    OK = 200
    CREATED = 201
    ACCEPTED = 202


POD_NAME_PATTERN = r"^[a-z0-9](?:[a-z0-9-]{0,61})[a-z0-9]?$"


# Source below based from: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/api_client.py
# Credit to original authors
# Original code Licensed under the Apache-2.0 license
# See the License file for the full license text.
def sanitize_for_serialization(obj: Any) -> Any:
    if obj is None:
        return None
    elif isinstance(obj, (float, bool, bytes, str, int)):
        return obj
    elif isinstance(obj, list):
        return [sanitize_for_serialization(sub_obj) for sub_obj in obj]
    elif isinstance(obj, tuple):
        return tuple(sanitize_for_serialization(sub_obj) for sub_obj in obj)
    elif isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat

    if isinstance(obj, dict):
        obj_dict = obj
    else:
        obj_dict = {
            obj.attribute_map[attr]: getattr(obj, attr)
            for attr, _ in obj.openapi_types.items()
            if getattr(obj, attr) is not None
        }

    return {key: sanitize_for_serialization(val) for key, val in obj_dict.items()}


async def wait_for_ready(
    kube_config: Any,
    pod_name: str,
    namespace: str,
    timeout: int = 60,
    stop_time: float = 0.1,
) -> Any:
    elapsed_time = 0.0
    while True:
        pod = await _read_pod_status(kube_config, pod_name, namespace)
        if pod["status"]["phase"] == "Running" or elapsed_time >= timeout:
            break
        await asyncio.sleep(stop_time)
        elapsed_time += stop_time
    if pod["status"]["phase"] != "Running":
        raise ValueError("Pod failed to start")
    else:
        return pod


def _create_ssl_context_and_headers(
    kube_config: Any,
) -> tuple[ssl.SSLContext, dict[str, str]]:
    ca_cert = kube_config.ssl_ca_cert
    ssl_context = ssl.create_default_context(cafile=ca_cert)
    headers = {}

    if "authorization" in kube_config.api_key:  ## Bearer token
        headers.update({"Authorization": kube_config.api_key["authorization"]})
    else:  ## ssl cafile and keyfile
        ssl_context.load_cert_chain(certfile=kube_config.cert_file, keyfile=kube_config.key_file)
    return ssl_context, headers


websocket_subprotocol_kubernetes_api = os.environ.get("WEBSOCKET_SUBPROTOCOL_KUBERNETES_API", "v4.channel.k8s.io")


async def pod_exec_stream(
    kube_config: Any,
    pod_name: str,
    namespace: str,
    command: list[str],
    container_name: str,
) -> AsyncGenerator[tuple[int, str, Optional[int]], None]:
    api_server_url = kube_config.host

    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    additional_headers = [(key, value) for key, value in headers.items()]

    # websocket
    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}/exec"
    params = {
        "container": container_name,
        "stdin": "false",
        "stdout": "true",
        "stderr": "true",
        "tty": "false",
    }
    if not command:
        raise ValueError("command must not be empty list")
    command_query_string = "&".join(f"command={urllib.parse.quote_plus(cmd)}" for cmd in command)
    query_string = "&".join(f"{key}={urllib.parse.quote_plus(value)}" for key, value in params.items() if value)
    websocket_url = f"{url}?{command_query_string}&{query_string}".replace("https://", "wss://")
    subprotocols = [Subprotocol(websocket_subprotocol_kubernetes_api)]

    async with connect(
        websocket_url,
        ssl=ssl_context,
        additional_headers=additional_headers,
        subprotocols=subprotocols,
    ) as ws:
        try:
            while True:
                message = await ws.recv(decode=False)
                if len(message) < 2:
                    continue
                channel = int(message[0])
                msg_content = message[1:]
                if isinstance(msg_content, bytes):
                    content = msg_content.decode("utf-8")
                else:
                    content = msg_content

                returncode = None
                if channel == StreamChannel.ERROR_CHANNEL:  # ERROR_CHANNEL
                    error_info = yaml.safe_load(content)
                    if error_info["status"] == "Success":
                        returncode = 0
                    else:
                        returncode = int(error_info["details"]["causes"][0]["message"])

                yield (channel, content, returncode)

        except ConnectionClosed as e:
            logging.info(f"Websocket connection closed: {e}")
        except asyncio.CancelledError:
            payload = json.dumps({"signal": "SIGKILL"}).encode("utf-8")
            frame = b"\x03" + payload
            await ws.send(frame)

            logging.info("send SIGKILL")
            raise


async def _read_pod_status(kube_config: Any, pod_name: str, namespace: str) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update({"Accept": "application/json"})

    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}/status"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.get(url, headers=headers)
        if response.status_code == HttpStatusCode.OK:
            return response.json()
        else:
            logging.info("Failed to get pod status: {response.status_code}")
            response.raise_for_status()


async def get_pod_logs(
    kube_config: Any,
    pod_name: str,
    namespace: str,
    container_name: str,
) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update({"Accept": "application/json", "Content-Type": "application/json"})

    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}/log?container={container_name}"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.get(url, headers=headers)
        if response.status_code == HttpStatusCode.OK:
            return response.text
        else:
            logging.info("Failed to get pod status: {response.status_code}")
            response.raise_for_status()


async def create_namespaced_corev1_resource(
    kube_config: Any, resource_spec: dict[str, Any], dry_run: bool = False
) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update(
        {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
    )

    namespace = resource_spec["metadata"]["namespace"]
    kind = str(resource_spec["kind"]).lower() + "s"
    url = f"{api_server_url}/api/v1/namespaces/{namespace}/{kind}"
    if dry_run:
        url += "?dryRun=All"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.post(url, headers=headers, json=resource_spec)
        if response.status_code == HttpStatusCode.CREATED:
            return response.json()
        else:
            logging.info("Failed to create pod: {response.status_code}")
            response.raise_for_status()


async def delete_namespaced_corev1_resource(kube_config: Any, resource: dict[str, Any]) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update({"Accept": "application/json"})

    kind = resource["kind"]
    name = resource["metadata"]["name"]
    namespace = resource["metadata"]["namespace"]

    url = f"{api_server_url}/api/v1/namespaces/{namespace}/{kind.lower() + 's'}/{name}"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.delete(url, headers=headers)
        if response.status_code in [HttpStatusCode.OK, HttpStatusCode.ACCEPTED]:
            return response.json()
        else:
            logging.info("Failed to get pod status: {response.status_code}")
            response.raise_for_status()


# Source below based from: https://github.com/microsoft/autogen/blob/main/python/packages/autogen-ext/src/autogen_ext/code_executors/_common.py
# Credit to original authors
# Original code Licensed under the MIT License.
# See the License file for the full license text.


@dataclass
class CommandLineCodeResult(CodeResult):
    """A code result class for command line code executor."""

    code_file: Optional[str]


# Raises ValueError if the file is not in the workspace
def get_file_name_from_content(code: str, workspace_path: Path) -> Optional[str]:
    first_line = code.split("\n")[0]
    # TODO - support other languages
    if first_line.startswith("# filename:"):
        filename = first_line.split(":")[1].strip()

        # Handle relative paths in the filename
        path = Path(filename)
        if not path.is_absolute():
            path = workspace_path / path
        path = path.resolve()
        # Throws an error if the file is not in the workspace
        relative = path.relative_to(workspace_path.resolve())
        return str(relative)

    return None


PYTHON_VARIANTS = ["python", "Python", "py"]


def lang_to_cmd(lang: str) -> str:
    if lang in PYTHON_VARIANTS:
        return "python"
    if lang.startswith("python") or lang in ["bash", "sh"]:
        return lang
    if lang in ["shell"]:
        return "sh"
    else:
        raise ValueError(f"Unsupported language: {lang}")


def silence_pip(code: str, lang: str) -> str:
    """Apply -qqq flag to pip install commands."""
    if lang == "python":
        regex = r"^! ?pip install"
    elif lang in ["bash", "shell", "sh", "pwsh", "powershell", "ps1"]:
        regex = r"^pip install"
    else:
        return code

    # Find lines that start with pip install and make sure "-qqq" flag is added.
    lines = code.split("\n")
    for i, line in enumerate(lines):
        # use regex to find lines that start with pip install.
        match = re.search(regex, line)
        if match is not None:
            if "-qqq" not in line:
                lines[i] = line.replace(match.group(0), match.group(0) + " -qqq")
    return "\n".join(lines)



import logging
import os
import re
import socket
from pathlib import Path

import pytest
from autogen_core import CancellationToken
from autogen_core.code_executor import CodeBlock
from autogen_kubernetes.code_executors import PodJupyterCodeExecutor, PodJupyterServer
from conftest import kubernetes_enabled, state_kubernetes_enabled

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def can_resolve_svc_fqdn() -> bool:
    try:
        socket.gethostbyname("kubernetes.default")
        return True
    except socket.error:
        return False


@pytest.mark.skipif(
    not state_kubernetes_enabled or not can_resolve_svc_fqdn(),
    reason="kubernetes not accessible",
)
@pytest.mark.asyncio
async def test_pod_default(generated_jupyter_pod_regex: str) -> None:
    async with PodJupyterServer() as jupyter_server:
        async with PodJupyterCodeExecutor(jupyter_server) as executor:
            code_result = await executor.execute_code_blocks(
                code_blocks=[
                    CodeBlock(language="python", code="print('Hello, World!')"),
                ],
                cancellation_token=CancellationToken(),
            )
            assert re.fullmatch(generated_jupyter_pod_regex, jupyter_server._pod_name) is not None
            assert code_result.exit_code == 0
            assert "Hello, World!" in code_result.output


@pytest.mark.skipif(
    not state_kubernetes_enabled or not can_resolve_svc_fqdn(),
    reason="kubernetes not accessible",
)
@pytest.mark.asyncio
async def test_pod_yaml_file() -> None:
    yaml_file_path = Path(os.path.dirname(__file__)) / Path("spec-files/test-jupyter.yaml")

    async with PodJupyterServer(
        pod_spec=yaml_file_path,
        secret_spec=yaml_file_path,
        service_spec=yaml_file_path,
    ) as jupyter_server:
        async with PodJupyterCodeExecutor(jupyter_server) as executor:
            code_result = await executor.execute_code_blocks(
                code_blocks=[
                    CodeBlock(language="python", code="print('Hello, World!')"),
                ],
                cancellation_token=CancellationToken(),
            )
            assert jupyter_server.connection_info.host == "test-jupyter.default.svc.cluster.local"
            assert jupyter_server.connection_info.port == 8888
            assert code_result.exit_code == 0
            assert "Hello, World!" in code_result.output



import asyncio
import base64
import datetime
import json
import re
import secrets
import time
import uuid
import warnings
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Type,
    Union,
    cast,
    runtime_checkable,
)
from urllib.parse import urlparse

import asyncio_atexit
import httpx
import kubernetes.client
import kubernetes.config
import yaml
from autogen_core import Component, ComponentBase
from httpx import HTTPStatusError
from kubernetes.client.models import (
    V1Container,
    V1ContainerPort,
    V1EnvFromSource,
    V1EnvVar,
    V1EnvVarSource,
    V1ObjectMeta,
    V1Pod,
    V1PodSpec,
    V1Secret,
    V1SecretEnvSource,
    V1Service,
    V1ServicePort,
    V1ServiceSpec,
    V1Volume,
    V1VolumeMount,
)
from pydantic import BaseModel, SecretStr
from typing_extensions import Self
from websockets.asyncio.client import ClientConnection, connect

from ._utils import (
    create_namespaced_corev1_resource,
    delete_namespaced_corev1_resource,
    get_pod_logs,
    sanitize_for_serialization,
    wait_for_ready,
)


class PodJupyterConnectionInfo(BaseModel):
    host: str
    port: Optional[int] = None
    token: Optional[SecretStr] = None


class PodJupyterClient:
    def __init__(self, connection_info: PodJupyterConnectionInfo):
        self._connection_info = connection_info
        self._http_client: httpx.AsyncClient = httpx.AsyncClient(headers=self._get_headers())

    def _get_headers(self) -> Dict[str, str]:
        if self._connection_info.token is None:
            return {}
        return {"Authorization": f"token {self._connection_info.token.get_secret_value()}"}

    def _get_api_base_url(self) -> str:
        port = f":{self._connection_info.port}" if self._connection_info.port else ""
        if urlparse(self._connection_info.host).scheme:
            api_server_url = self._connection_info.host
        else:
            api_server_url = f"http://{self._connection_info.host}"
        return f"{api_server_url}{port}"

    def _get_ws_base_url(self) -> str:
        api_server_url = self._get_api_base_url().replace("http://", "ws://").replace("https://", "wss://")
        return api_server_url

    async def list_kernel_specs(self) -> Dict[str, Dict[str, str]]:
        response = await self._http_client.get(
            f"{self._get_api_base_url()}/api/kernelspecs",
        )
        response.raise_for_status()
        return cast(Dict[str, Dict[str, str]], response.json())

    async def list_kernels(self) -> List[Dict[str, str]]:
        response = await self._http_client.get(f"{self._get_api_base_url()}/api/kernels")
        response.raise_for_status()
        return cast(List[Dict[str, str]], response.json())

    async def start_kernel(self, kernel_spec_name: str) -> str:
        """Start a new kernel asynchronously.

        Args:
            kernel_spec_name (str): Name of the kernel spec to start

        Returns:
            str: ID of the started kernel
        """
        response = await self._http_client.post(
            f"{self._get_api_base_url()}/api/kernels",
            json={"name": kernel_spec_name},
        )
        response.raise_for_status()
        data = response.json()
        return cast(str, data["id"])

    async def delete_kernel(self, kernel_id: str) -> None:
        response = await self._http_client.delete(
            f"{self._get_api_base_url()}/api/kernels/{kernel_id}",
        )
        response.raise_for_status()

    async def restart_kernel(self, kernel_id: str) -> None:
        response = await self._http_client.post(
            f"{self._get_api_base_url()}/api/kernels/{kernel_id}/restart",
        )
        response.raise_for_status()

    async def get_kernel_client(self, kernel_id: str) -> "JupyterKernelClient":
        ws_path = f"/api/kernels/{kernel_id}/channels"
        # Using websockets library for async websocket connections
        ws = await connect(
            uri=self._get_ws_base_url() + ws_path,
            additional_headers=self._get_headers(),
        )
        return JupyterKernelClient(ws)

    async def close(self) -> None:
        """Close the async session"""
        await self._http_client.aclose()


DEFAULT_COMMAND = [
    "/bin/bash",
    "-o",
    "pipefail",
    "-c",
    """
    mamba install --yes jupyter_kernel_gateway ipykernel && \
        mamba clean --all -f -y && \
        fix-permissions "${CONDA_DIR}" && \
        fix-permissions "/home/${NB_USER}"; \
    python -m jupyter kernelgateway --KernelGatewayApp.ip=0.0.0.0 \
        --JupyterApp.answer_yes=true \
        --JupyterWebsocketPersonality.list_kernels=true;
    """,
]
DEFAULT_PORT = 8888


class PodJupyterServerConfig(BaseModel):
    """Configuration for PodJupyterServer"""

    image: Optional[str] = None
    pod_name: Optional[str] = None
    command: Optional[list[str]] = None
    args: Optional[list[str]] = None
    timeout: int = 60
    workspace_path: Union[Path, str] = Path("/workspace")
    namespace: str = "default"
    volume: Union[dict[str, Any], str, Path, Type[V1Volume], None] = None
    pod_spec: Union[dict[str, Any], str, Path, Type[V1Pod], None] = None
    auto_remove: bool = True
    port: int = DEFAULT_PORT
    service_spec: Union[dict[str, Any], str, Path, Type[V1Service], None] = None
    token: Optional[SecretStr] = None
    secret_spec: Union[dict[str, Any], str, Path, Type[V1Secret], SecretStr, None] = None
    """when secret spec is provided as SecretStr, It should be a json format"""
    kube_config_file: Union[Path, str, None] = None


class PodJupyterServer(Component[PodJupyterServerConfig], ComponentBase[BaseModel]):
    class GenerateToken:
        pass

    component_type = "jupyter_server"
    component_version = 1
    component_description = "jupyter server component as kubernetes pod"
    component_config_schema = PodJupyterServerConfig
    component_provider_override = "autogen_kubernetes.code_executors.PodJupyterServer"

    def __init__(
        self,
        *,
        image: Optional[str] = None,
        pod_name: Optional[str] = None,
        command: Optional[list[str]] = DEFAULT_COMMAND,
        args: Optional[list[str]] = None,
        timeout: int = 60,
        workspace_path: Union[Path, str] = "/workspace",
        namespace: str = "default",
        volume: Union[dict[str, Any], str, Path, Type[V1Volume], None] = None,
        pod_spec: Union[dict[str, Any], str, Path, Type[V1Pod], None] = None,
        auto_remove: bool = True,
        port: int = DEFAULT_PORT,
        service_spec: Union[dict[str, Any], str, Path, Type[V1Service], None] = None,
        token: Optional[Union[str, GenerateToken]] = None,
        secret_spec: Union[dict[str, Any], str, Path, Type[V1Secret], None] = None,
        kube_config_file: Union[Path, str, None] = None,
    ):
        """Start a Jupyter kernel gateway server in a Kubernetes container.

        Args:
            image (Optional[str], optional): jupyter server docker image. Defaults to None and converted to quay.io/jupyter/docker-stacks-foundation.
            pod_name (Optional[str], optional): jupyter server pod name. Defaults to None and will be converted to auto generated name.
            command (Optional[list[str]], optional): jupyter server container command. Defaults to DEFAULT_COMMAND.
            args (Optional[list[str]], optional): jupyter server container argument. Defaults to None.
            timeout (int, optional): wait time in secondes for the pod to be running phase. Defaults to 60.
            workspace_path (str, optional): default workspace directory. Defaults to "/workspace".
            namespace (str, optional): namespace for jupyter server pod. Defaults to "default".
            volume (Union[dict[str, Any], str, Path, Type[V1Volume], None], optional): The volume for the jupyter server pod.
                Supports the formats of a dictionary, a YAML/json string, a YAML file path, and kubernetes `V1Volume` model format.
                Must conform to the kubernetes `V1Volume` model format.
                Must have appropriate access mode(such as ReadWriteMany, ReadWriteOnce, ReadWriteOncePod, in case of PersistentVolumeClaim)
                If None, no volume attached to code executor pod. Defaults to None.
            pod_spec (Union[dict[str, Any], str, Path, Type[V1Pod], None], optional): Custom pod specification for jupyter server.
                Must contain a container which name is "autogen-executor" for execution for codes.
                Supports the formats of a dictionary, a YAML/json string, a YAML file path, and a kubernetes V1Pod model.
                Must conform to the kubernetes `V1Pod` model format.
                It has priority than above parameters.
                If None, will use above parameters to create code executor pod. Defaults to None.
            auto_remove (bool, optional): If true, will automatically remove the
                container when remove is called, when the context manager exits or when
                the Python process exits with atext. Defaults to True.
            port (int, optional): port number for jupyter server. Defaults to DEFAULT_PORT.
            service_spec (Union[dict[str, Any], str, Path, Type[V1Service], None], optional): Custom service specificatin for jupyter server.
                Supports the formats of a dictionary, a YAML/json string, a YAML file path, and kubernetes `V1Service` model format.
                Must conform to the kubernetes `V1Service` model format.
                The first port should be the name `jupyter` otherwise PodJupyterClient cannot use the port.
                It has priority than port parameter.
                Defaults to None.
            token (Optional[Union[str, GenerateToken]], optional): token string. Defaults to None and will be converted to auto generated token.
            secret_spec (Union[dict[str, Any], str, Path, Type[V1Secret], None], optional): Custom secret specificatin for jupyter server.
                Supports the formats of a dictionary, a YAML/json string, a YAML file path, and kubernetes `V1Secret` model format.
                Must conform to the kubernetes `V1Secret` model format.
                It is possible to add data `KG_AUTH_TOKEN` to use custom authentication token for jupyter kernel gateway server.
                It has priority than token parameter.
                Defaults to None.
            kube_config_file (Union[Path, str, None], optional): kubernetes configuration file(kubeconfig) path.
                If None, will use `KUBECONFIG` environment variables or service account token(incluster config).
                Service account must have at least those namespaced permissions below.
                [
                {
                    "resource": "pods", "verb": ["get", "create", "delete"]
                },
                {
                    "resource": "pods/status", "verb": ["get"]
                },
                {
                    "resource": "pods/log", "verb": ["get"]
                },
                {
                    "resource": "services", "verb": ["get", "create", "delete"]
                },
                {
                    "resource": "secrets", "verb": ["get", "create", "delete"]
                },
                ]
        """
        # Generate container name if not provided
        self._pod_name = pod_name or f"autogen-jupyter-{uuid.uuid4()}"
        self._secret_name = self._pod_name + "-secret"

        if not re.fullmatch(r"^[a-z0-9](?:[a-z0-9-]{0,54})[a-z0-9]?$", self._pod_name):
            raise ValueError(
                "Pod name validation failed: pod name must start and end with lower alphanumeric characters with length 56, "
                "and contain only lowercase alphanumeric or dash('-') characters."
            )

        # Determine and prepare Docker image
        self._image = image or "quay.io/jupyter/docker-stacks-foundation"

        # Set up authentication token
        if token is None:
            token = PodJupyterServer.GenerateToken()
        self._token = secrets.token_hex(32) if isinstance(token, PodJupyterServer.GenerateToken) else token

        ## kubeconfig
        if isinstance(kube_config_file, str):
            kube_config_file = Path(kube_config_file)

        if kube_config_file is None:  ## configuration from default kubeconfig or incluster
            kubernetes.config.load_config()  # type: ignore
        else:
            kubernetes.config.load_config(config_file=kube_config_file)  # type: ignore
        self._kube_config = kubernetes.client.Configuration.get_default_copy()  # type: ignore
        self._kube_config_file = kube_config_file

        self._port = port
        self._container_name = "autogen-executor"
        self._namespace = namespace
        self._timeout = timeout
        self._auto_remove = auto_remove
        self._command = command
        self._args = args
        ## workspace
        if isinstance(workspace_path, str):  ## path string to Path
            workspace_path = Path(workspace_path)

        if not workspace_path.is_absolute():  ## validate workspace_path
            raise ValueError("Not supports workspace path as relative path")

        self._workspace_path: Path = workspace_path

        ## volume
        volume_json = None
        if isinstance(volume, str):
            try:
                if Path(volume).exists():  # YAML file path string
                    volume = Path(volume)
            except OSError:
                pass

        if isinstance(volume, Path):  # YAML file to dict
            volume_json = yaml.safe_load(volume.read_text(encoding="utf-8"))
        elif isinstance(volume, str):  # YAML string to dict
            volume_json = yaml.safe_load(volume)
        elif isinstance(volume, V1Volume):
            volume_json = sanitize_for_serialization(volume)
        elif isinstance(volume, dict):
            volume_json = volume

        self._volume = volume_json

        ## set default resources
        self._secret: Optional[dict[str, Any]] = None
        self._define_resources_spec()

        if pod_spec is not None:
            self._pod = self._read_from_resource_spec("Pod", pod_spec)
        if secret_spec is not None:
            self._secret = self._read_from_resource_spec("Secret", secret_spec)
            if "KG_AUTH_TOKEN" in self._secret["data"]:
                self._token = base64.b64decode(self._secret["data"]["KG_AUTH_TOKEN"]).decode("utf-8")
            elif "KG_AUTH_TOKEN" in self._secret["stringData"]:
                self._token = self._secret["stringData"]["KG_AUTH_TOKEN"]
        if service_spec is not None:
            self._service = self._read_from_resource_spec("Service", service_spec)
            has_jupyter_port = False
            for p in self._service["spec"]["ports"]:
                if p["name"] == "jupyter":
                    self._port = p["port"]
                    has_jupyter_port = True
                    break
            if not has_jupyter_port:
                raise ValueError(
                    "custom service should have a port named `jupyter`. add a port named `jupyter` in service_spec"
                )

        self._running = False

    def _read_from_resource_spec(self, kind: str, spec: Any) -> dict[str, Any]:
        types = {"Pod": V1Pod, "Secret": V1Secret, "Service": V1Service}

        if isinstance(spec, str):  # YAML file path string
            try:
                if Path(spec).exists():
                    spec = Path(spec)
            except OSError:
                pass

        resource_dict = {}
        if isinstance(spec, Path):  # YAML file to dict
            for obj in yaml.safe_load_all(spec.read_text(encoding="utf-8")):
                if obj["kind"] == kind:
                    resource_dict = obj
            if not resource_dict:
                raise ValueError(f"spec file {str(spec)} has no manifest for kind {kind}")
        elif isinstance(spec, str):  # YAML string to dict
            resource_dict = yaml.safe_load(spec)
        elif isinstance(spec, types[kind]):
            resource_dict = sanitize_for_serialization(spec)  # type: ignore
        elif isinstance(spec, dict):
            resource_dict = spec

        return resource_dict

    def _define_resources_spec(self) -> None:
        label = {"autogen-kubernetes/jupyter-server": self._pod_name}
        pod = V1Pod(
            kind="Pod",
            metadata=V1ObjectMeta(
                name=self._pod_name,
                namespace=self._namespace,
                labels=label,
            ),
        )

        executor_container = V1Container(
            args=self._args,
            command=self._command,
            name=self._container_name,
            image=self._image,
            env=[
                V1EnvVar(name="KG_PORT", value=str(self._port)),
            ],
            ports=[V1ContainerPort(container_port=self._port, name="jupyter")],
            working_dir=str(self._workspace_path),
        )
        if self._token:
            executor_container.env_from = [
                V1EnvFromSource(secret_ref=V1SecretEnvSource(name=self._secret_name)),
            ]

            secret = V1Secret(
                kind="Secret",
                metadata=V1ObjectMeta(name=self._secret_name, namespace=self._namespace, labels=label),
                string_data={"KG_AUTH_TOKEN": self._token},
            )
            self._secret = sanitize_for_serialization(secret)

        pod_spec = V1PodSpec(
            automount_service_account_token=False,
            containers=[executor_container],
        )

        if self._volume:
            executor_container.volume_mounts = [
                V1VolumeMount(mount_path=str(self._workspace_path), name=self._volume["name"])
            ]
            pod_spec.volumes = [V1Volume(**self._volume)]

        pod.spec = pod_spec
        self._pod = sanitize_for_serialization(pod)

        service = V1Service(
            kind="Service",
            metadata=V1ObjectMeta(
                name=self._pod_name,
                namespace=self._namespace,
                labels=label,
            ),
            spec=V1ServiceSpec(
                type="ClusterIP",
                selector=label,
                ports=[
                    V1ServicePort(
                        name="jupyter",
                        protocol="TCP",
                        port=self._port,
                        target_port="jupyter",
                    )
                ],
            ),
        )
        self._service = sanitize_for_serialization(service)

    @property
    def connection_info(self) -> PodJupyterConnectionInfo:
        return PodJupyterConnectionInfo(
            host=f"{self._service['metadata']['name']}.{self._service['metadata']['namespace']}.svc.cluster.local",
            port=self._port,
            token=SecretStr(self._token) if self._token else None,
        )

    async def remove(self, spec: Optional[dict[str, Any]]) -> None:
        if spec is None:
            return
        try:
            await delete_namespaced_corev1_resource(self._kube_config, spec)
        except HTTPStatusError:
            pass

    async def stop(self) -> None:
        await asyncio.gather(*[self.remove(spec) for spec in (self._pod, self._secret, self._service)])
        self._running = False

    async def start(self) -> None:
        for spec in [self._secret, self._pod, self._service]:
            if spec is None:
                continue
            await create_namespaced_corev1_resource(self._kube_config, spec)

        self._pod = await wait_for_ready(
            self._kube_config,
            self._pod["metadata"]["name"],
            self._pod["metadata"]["namespace"],
            self._timeout,
        )

        async def cleanup() -> None:
            await self.stop()
            asyncio_atexit.unregister(cleanup)  # type: ignore

        if self._auto_remove:
            asyncio_atexit.register(cleanup)  # type: ignore

        # Check if the container is running
        if self._pod["status"]["phase"] != "Running":
            logs_str = await get_pod_logs(
                self._kube_config,
                self._pod["metadata"]["name"],
                self._pod["metadata"]["namespace"],
                self._container_name,
            )
            raise ValueError(f"Failed to start container from image {self._image}. Logs: {logs_str}")

        self._running = True

    async def get_client(self) -> PodJupyterClient:
        return PodJupyterClient(self.connection_info)

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.stop()

    def _to_config(self) -> PodJupyterServerConfig:
        """(Experimental) Convert the component to a config object"""

        return PodJupyterServerConfig(
            image=self._image,
            pod_name=self._pod_name,
            timeout=self._timeout,
            workspace_path=self._workspace_path,
            namespace=self._namespace,
            volume=self._volume,
            pod_spec=self._pod,
            kube_config_file=self._kube_config_file,
            auto_remove=self._auto_remove,
            port=self._port,
            command=self._command,
            service_spec=self._service,
            token=SecretStr(self._token),
            secret_spec=SecretStr(str(self._secret)),
        )

    @classmethod
    def _from_config(cls, config: PodJupyterServerConfig) -> Self:
        """(Experimental) Create a component from a config object"""

        return cls(
            image=config.image,
            pod_name=config.pod_name,
            command=config.command,
            args=config.args,
            timeout=config.timeout,
            workspace_path=config.workspace_path,
            namespace=config.namespace,
            volume=config.volume,
            pod_spec=config.pod_spec,
            auto_remove=config.auto_remove,
            port=config.port,
            service_spec=config.service_spec,
            token=(config.token.get_secret_value() if config.token is not None else config.token),
            secret_spec=(
                config.secret_spec
                if not isinstance(config.secret_spec, SecretStr)
                else json.loads(config.secret_spec.get_secret_value())
            ),
            kube_config_file=config.kube_config_file,
        )


# Source below based from: https://github.com/microsoft/autogen/blob/main/python/packages/autogen-ext/src/autogen_ext/code_executors/docker_jupyter/_jupyter_server.py
# Credit to original authors
# Original code Licensed under the MIT License.
# See the License file for the full license text.


@dataclass
class DataItem:
    mime_type: str
    data: str


@dataclass
class ExecutionResult:
    is_ok: bool
    output: str
    data_items: List[DataItem]


class JupyterKernelClient:
    """An asynchronous client for communicating with a Jupyter kernel."""

    def __init__(self, websocket: ClientConnection) -> None:
        self._session_id = uuid.uuid4().hex
        self._websocket = websocket

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.stop()

    async def stop(self) -> None:
        await self._websocket.close()

    async def _send_message(self, *, content: Dict[str, Any], channel: str, message_type: str) -> str:
        timestamp = datetime.datetime.now().isoformat()
        message_id = uuid.uuid4().hex
        message = {
            "header": {
                "username": "autogen",
                "version": "5.0",
                "session": self._session_id,
                "msg_id": message_id,
                "msg_type": message_type,
                "date": timestamp,
            },
            "parent_header": {},
            "channel": channel,
            "content": content,
            "metadata": {},
            "buffers": {},
        }
        await self._websocket.send(json.dumps(message))
        return message_id

    async def _receive_message(self, timeout_seconds: Optional[float]) -> Optional[Dict[str, Any]]:
        try:
            if timeout_seconds is not None:
                data = await asyncio.wait_for(self._websocket.recv(), timeout=timeout_seconds)
            else:
                data = await self._websocket.recv()
            if isinstance(data, bytes):
                return cast(Dict[str, Any], json.loads(data.decode("utf-8")))
            return cast(Dict[str, Any], json.loads(data))
        except asyncio.TimeoutError:
            return None

    async def wait_for_ready(self, timeout_seconds: Optional[float] = None) -> bool:
        message_id = await self._send_message(content={}, channel="shell", message_type="kernel_info_request")
        while True:
            message = await self._receive_message(timeout_seconds)
            # This means we timed out with no new messages.
            if message is None:
                return False
            if (
                message.get("parent_header", {}).get("msg_id") == message_id
                and message["msg_type"] == "kernel_info_reply"
            ):
                return True

    async def execute(self, code: str, timeout_seconds: Optional[float] = None) -> ExecutionResult:
        message_id = await self._send_message(
            content={
                "code": code,
                "silent": False,
                "store_history": True,
                "user_expressions": {},
                "allow_stdin": False,
                "stop_on_error": True,
            },
            channel="shell",
            message_type="execute_request",
        )

        text_output: List[str] = []
        data_output: List[DataItem] = []
        while True:
            message = await self._receive_message(timeout_seconds)
            if message is None:
                return ExecutionResult(
                    is_ok=False,
                    output="ERROR: Timeout waiting for output from code block.",
                    data_items=[],
                )

            # Ignore messages that are not for this execution.
            if message.get("parent_header", {}).get("msg_id") != message_id:
                continue

            msg_type = message["msg_type"]
            content = message["content"]
            if msg_type in ["execute_result", "display_data"]:
                for data_type, data in content["data"].items():
                    if data_type == "text/plain":
                        text_output.append(data)
                    elif data_type.startswith("image/") or data_type == "text/html":
                        data_output.append(DataItem(mime_type=data_type, data=data))
                    else:
                        text_output.append(json.dumps(data))
            elif msg_type == "stream":
                text_output.append(content["text"])
            elif msg_type == "error":
                # Output is an error.
                return ExecutionResult(
                    is_ok=False,
                    output=f"ERROR: {content['ename']}: {content['evalue']}\n{content['traceback']}",
                    data_items=[],
                )
            if msg_type == "status" and content["execution_state"] == "idle":
                break
        return ExecutionResult(
            is_ok=True,
            output="\n".join([str(output) for output in text_output]),
            data_items=data_output,
        )



from ._jupyter_server import (
    PodJupyterConnectionInfo,
    PodJupyterServer,
    PodJupyterServerConfig,
)
from ._kubernetes_code_executor import (
    PodCommandLineCodeExecutor,
    PodCommandLineCodeExecutorConfig,
)
from ._kubernetes_jupyter import PodJupyterCodeExecutor, PodJupyterCodeExecutorConfig

__all__ = [
    "PodCommandLineCodeExecutor",
    "PodCommandLineCodeExecutorConfig",
    "PodJupyterCodeExecutor",
    "PodJupyterCodeExecutorConfig",
    "PodJupyterServer",
    "PodJupyterServerConfig",
    "PodJupyterConnectionInfo",
]



import asyncio
import base64
import json
import os
import tempfile
import uuid
import warnings
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import List, Optional, Union

import httpx
from autogen_core import CancellationToken, Component
from autogen_core.code_executor import CodeBlock, CodeExecutor, CodeResult
from pydantic import BaseModel, ConfigDict
from typing_extensions import Self

from ._jupyter_server import (
    JupyterKernelClient,
    PodJupyterClient,
    PodJupyterConnectionInfo,
    PodJupyterServer,
)
from ._utils import silence_pip


# Source below based from: https://github.com/microsoft/autogen/blob/main/python/packages/autogen-ext/src/autogen_ext/code_executors/docker_jupyter/_docker_jupyter.py
# Credit to original authors
# Original code Licensed under the MIT License.
# See the License file for the full license text.
@dataclass
class PodJupyterCodeResult(CodeResult):
    """(Experimental) A code result class for IPython code executor."""

    output_files: list[Path]


class PodJupyterCodeExecutorConfig(BaseModel):
    """Configuration for JupyterCodeExecutor"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    jupyter_server: Union[PodJupyterServer, PodJupyterConnectionInfo]
    kernel_name: str = "python3"
    timeout: int = 60
    output_dir: Optional[Union[Path, str]] = None


class PodJupyterCodeExecutor(CodeExecutor, Component[PodJupyterCodeExecutorConfig]):
    """(Experimental) A code executor class that executes code statefully using
    a Jupyter server supplied to this class.

    Each execution is stateful and can access variables created from previous
    executions in the same session.

    Args:
        jupyter_server (Union[JupyterConnectable, JupyterConnectionInfo]): The Jupyter server to use.
        kernel_name (str): The kernel name to use. Make sure it is installed.
            By default, it is "python3".
        timeout (int): The timeout for code execution, by default 60.
        output_dir (Path): The directory to save output files, by default None.

    Example of using it directly:

    .. code-block:: python

        import asyncio
        from autogen_core import CancellationToken
        from autogen_core.code_executor import CodeBlock
        from autogen_kubernetes.code_executors import PodJupyterCodeExecutor, PodJupyterServer


        async def main() -> None:
            async with PodJupyterServer() as jupyter_server:
                async with PodJupyterCodeExecutor(jupyter_server=jupyter_server) as executor:
                    code_blocks = [CodeBlock(code="print('hello world!')", language="python")]
                    code_result = await executor.execute_code_blocks(code_blocks, cancellation_token=CancellationToken())
                    print(code_result)


        asyncio.run(main())

    Example of using it with your own jupyter image:

    .. code-block:: python

        import asyncio
        from autogen_core import CancellationToken
        from autogen_core.code_executor import CodeBlock
        from autogen_kubernetes.code_executors import PodJupyterCodeExecutor, PodJupyterServer


        async def main() -> None:
            async with PodJupyterServer(image="your_custom_images_name", port=8888) as jupyter_server:
                async with PodJupyterCodeExecutor(jupyter_server=jupyter_server) as executor:
                    code_blocks = [CodeBlock(code="print('hello world!')", language="python")]
                    code_result = await executor.execute_code_blocks(code_blocks, cancellation_token=CancellationToken())
                    print(code_result)


        asyncio.run(main())

    Example of using it with :class:`~autogen_ext.tools.code_execution.PythonCodeExecutionTool`:

    .. code-block:: python

        import asyncio
        from autogen_agentchat.agents import AssistantAgent
        from autogen_kubernetes.code_executors import PodJupyterCodeExecutor, PodJupyterServer
        from autogen_ext.models.openai import OpenAIChatCompletionClient
        from autogen_ext.tools.code_execution import PythonCodeExecutionTool


        async def main() -> None:
            async with PodJupyterServer() as jupyter_server:
                async with PodJupyterCodeExecutor(jupyter_server=jupyter_server) as executor:
                    tool = PythonCodeExecutionTool(executor)
                    model_client = OpenAIChatCompletionClient(model="gpt-4o")
                    agent = AssistantAgent("assistant", model_client=model_client, tools=[tool])
                    result = await agent.run(task="What is the 10th Fibonacci number? Use Python to calculate it.")
                    print(result)


        asyncio.run(main())

    Example of using it inside a :class:`~autogen_agentchat.agents._code_executor_agent.CodeExecutorAgent`:

    .. code-block:: python

        import asyncio
        from autogen_agentchat.agents import CodeExecutorAgent
        from autogen_agentchat.messages import TextMessage
        from autogen_kubernetes.code_executors import PodJupyterCodeExecutor, PodJupyterServer
        from autogen_core import CancellationToken


        async def main() -> None:
            async with PodJupyterServer() as jupyter_server:
                async with PodJupyterCodeExecutor(jupyter_server=jupyter_server) as executor:
                    code_executor_agent = CodeExecutorAgent("code_executor", code_executor=executor)
                    task = TextMessage(
                        content='''Here is some code
                ```python
                print('Hello world')
                ```
                ''',
                        source="user",
                    )
                    response = await code_executor_agent.on_messages([task], CancellationToken())
                    print(response.chat_message)


        asyncio.run(main())

    """

    component_config_schema = PodJupyterCodeExecutorConfig
    component_provider_override = "autogen_kubernetes.code_executors.PodJupyterCodeExecutor"

    def __init__(
        self,
        jupyter_server: Union[PodJupyterServer, PodJupyterConnectionInfo],
        kernel_name: str = "python3",
        timeout: int = 60,
        output_dir: Path | None = None,
    ):
        if timeout < 1:
            raise ValueError("Timeout must be greater than or equal to 1.")

        if isinstance(jupyter_server, PodJupyterServer):
            self._connection_info = jupyter_server.connection_info
        elif isinstance(jupyter_server, PodJupyterConnectionInfo):
            self._connection_info = jupyter_server
        else:
            raise ValueError("jupyter_server must be a PodJupyterServer or PodJupyterConnectionInfo.")

        self._output_dir = output_dir
        if not self._output_dir:
            with tempfile.TemporaryDirectory() as temp_dir:
                self._output_dir = Path(temp_dir)
                self._output_dir.mkdir(exist_ok=True)

        self._jupyter_client = PodJupyterClient(self._connection_info)

        self._kernel_name = kernel_name
        self._timeout = timeout
        self._async_jupyter_kernel_client: Optional[JupyterKernelClient] = None
        self._kernel_id: Optional[str] = None

    async def _ensure_async_kernel_client(self) -> JupyterKernelClient:
        """Ensure that an async kernel client exists and return it."""
        if self._kernel_id is None:
            await self.start()
            assert self._kernel_id is not None
        if self._async_jupyter_kernel_client is None:
            self._async_jupyter_kernel_client = await self._jupyter_client.get_kernel_client(self._kernel_id)
        return self._async_jupyter_kernel_client

    async def execute_code_blocks(
        self, code_blocks: List[CodeBlock], cancellation_token: CancellationToken
    ) -> PodJupyterCodeResult:
        """(Experimental) Execute a list of code blocks and return the result.

        This method executes a list of code blocks as cells in the Jupyter kernel.
        See: https://jupyter-client.readthedocs.io/en/stable/messaging.html
        for the message protocol.

        Args:
            code_blocks (List[CodeBlock]): A list of code blocks to execute.

        Returns:
            PodJupyterCodeResult: The result of the code execution.
        """
        kernel_client = await self._ensure_async_kernel_client()
        # Wait for kernel to be ready using async client
        is_ready = await kernel_client.wait_for_ready(timeout_seconds=self._timeout)
        if not is_ready:
            return PodJupyterCodeResult(exit_code=1, output="ERROR: Kernel not ready", output_files=[])

        outputs: List[str] = []
        output_files: List[Path] = []
        for code_block in code_blocks:
            code = silence_pip(code_block.code, code_block.language)
            # Execute code using async client
            exec_task = asyncio.create_task(kernel_client.execute(code, timeout_seconds=self._timeout))
            cancellation_token.link_future(exec_task)
            result = await exec_task
            if result.is_ok:
                outputs.append(result.output)
                for data in result.data_items:
                    if data.mime_type == "image/png":
                        path = self._save_image(data.data)
                        outputs.append(path)
                        output_files.append(Path(path))
                    elif data.mime_type == "text/html":
                        path = self._save_html(data.data)
                        outputs.append(path)
                        output_files.append(Path(path))
                    else:
                        outputs.append(json.dumps(data.data))
            else:
                existing_output = "\n".join([str(output) for output in outputs])
                return PodJupyterCodeResult(
                    exit_code=1,
                    output=existing_output + "\nERROR: " + result.output,
                    output_files=output_files,
                )
        return PodJupyterCodeResult(
            exit_code=0,
            output="\n".join([str(output) for output in outputs]),
            output_files=output_files,
        )

    async def restart(self) -> None:
        """(Experimental) Restart a new session."""
        # Use async client to restart kernel
        if self._kernel_id is not None:
            await self._jupyter_client.restart_kernel(self._kernel_id)
        # Reset the clients to force recreation
        if self._async_jupyter_kernel_client is not None:
            await self._async_jupyter_kernel_client.stop()
            self._async_jupyter_kernel_client = None

    async def _wait_for_server(self) -> None:
        while True:
            try:
                response = await self._jupyter_client.list_kernel_specs()
                assert "kernelspecs" in response
                break
            except (
                httpx.ConnectError,
                httpx.HTTPStatusError,
                AssertionError,
                json.decoder.JSONDecodeError,
            ):
                await asyncio.sleep(1)

    async def start(self) -> None:
        """(Experimental) Start a new session."""
        try:
            await asyncio.wait_for(self._wait_for_server(), timeout=self._timeout)
            available_kernels = await self._jupyter_client.list_kernel_specs()
            if self._kernel_name not in available_kernels["kernelspecs"]:
                raise ValueError(f"Kernel {self._kernel_name} is not installed.")
            self._kernel_id = await self._jupyter_client.start_kernel(self._kernel_name)
        except asyncio.TimeoutError:
            warnings.warn(
                f"jupyter server not accessible after connection tries for {self._timeout} seconds. close context.",
                stacklevel=2,
            )
            raise

    def _save_image(self, image_data_base64: str) -> str:
        """Save image data to a file."""
        image_data = base64.b64decode(image_data_base64)
        filename = f"{uuid.uuid4().hex}.png"
        path = os.path.join(str(self._output_dir), filename)
        with open(path, "wb") as f:
            f.write(image_data)
        return os.path.abspath(path)

    def _save_html(self, html_data: str) -> str:
        """Save html data to a file."""
        filename = f"{uuid.uuid4().hex}.html"
        path = os.path.join(str(self._output_dir), filename)
        with open(path, "w", encoding="utf-8") as f:
            f.write(html_data)
        return os.path.abspath(path)

    async def stop(self) -> None:
        """Stop the kernel."""
        if self._kernel_id is not None:
            await self._jupyter_client.delete_kernel(self._kernel_id)
        if self._async_jupyter_kernel_client is not None:
            await self._async_jupyter_kernel_client.stop()
            self._async_jupyter_kernel_client = None
        await self._jupyter_client.close()

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.stop()

    def _to_config(self) -> PodJupyterCodeExecutorConfig:
        """(Experimental) Convert the component to a config object"""

        return PodJupyterCodeExecutorConfig(
            jupyter_server=self._connection_info,
            kernel_name=self._kernel_name,
            timeout=self._timeout,
            output_dir=self._output_dir,
        )

    @classmethod
    def _from_config(cls, config: PodJupyterCodeExecutorConfig) -> Self:
        """(Experimental) Create a component from a config object"""

        return cls(
            jupyter_server=config.jupyter_server,
            kernel_name=config.kernel_name,
            timeout=config.timeout,
            output_dir=(config.output_dir if not isinstance(config.output_dir, str) else Path(config.output_dir)),
        )
